﻿// MidiStringPlayer.cs by Charles Petzold, January 2010
using System;
using System.Collections.Generic;
using System.Windows;
using System.Windows.Threading;
using NAudio.Midi;

namespace Petzold.Midi
{
    public class MidiStringPlayer : FrameworkElement
    {
        MidiOut midiOut;
        DispatcherTimer tmr;
        string midiString;
        List<MidiChannel> midiChannels = new List<MidiChannel>();
        List<MidiChannel> removeChannelList = new List<MidiChannel>();
        DateTime shutDownTime;
        bool haveSetWindowCloseHandler = false;

        public static readonly DependencyProperty MidiStringProperty =
            DependencyProperty.Register("MidiString",
                typeof(string),
                typeof(MidiStringPlayer));

        public static readonly DependencyProperty PlayOnLoadProperty =
            DependencyProperty.Register("PlayOnLoad",
                typeof(bool),
                typeof(MidiStringPlayer),
                new PropertyMetadata(false));

        private static readonly DependencyPropertyKey IsPlayingKey =
            DependencyProperty.RegisterReadOnly("IsPlaying",
                typeof(bool),
                typeof(MidiStringPlayer),
                new PropertyMetadata(false));

        public static readonly DependencyProperty IsPlayingProperty =
            IsPlayingKey.DependencyProperty;

        public event EventHandler Ended;
        public event MidiStringPlayerFailedEventHandler Failed;

        public MidiStringPlayer()
        {
            Loaded += OnLoaded;
        }

        public string MidiString
        {
            set { SetValue(MidiStringProperty, value); }
            get { return (string)GetValue(MidiStringProperty); }
        }

        public bool PlayOnLoad
        {
            set { SetValue(PlayOnLoadProperty, value); }
            get { return (bool)GetValue(PlayOnLoadProperty); }
        }

        public bool IsPlaying
        {
            protected set { SetValue(IsPlayingKey, value); }
            get { return (bool)GetValue(IsPlayingProperty); }
        }

        public bool Play()
        {
            if (!haveSetWindowCloseHandler)
            {
                Application.Current.MainWindow.Closed += OnWindowClosed;
                haveSetWindowCloseHandler = true;
            }

            if (IsPlaying)
                return false;

            if (String.IsNullOrEmpty(MidiString))
                return false;

            // Set private fields to default values
            midiString = MidiString;
            midiChannels.Clear();
            shutDownTime = DateTime.MaxValue;

            int startIndex = 0;
            int channel = 0;
            bool hasPercussionTrack = false;

            try
            {
                midiOut = new MidiOut(0);
            }
            catch (Exception exc)
            {
                OnFail(startIndex, "Cannot open MidiOut: " + exc.Message);
                return false;
            }

            DateTime startTime = DateTime.Now;

            do
            {
                if (HasContent(midiString, startIndex))
                {
                    // Check for percussion track
                    int endIndex = midiString.IndexOf('|', startIndex);

                    if (endIndex == -1)
                        endIndex = midiString.Length;

                    if (midiString.IndexOf('P', startIndex, endIndex - startIndex) != -1)
                    {
                        if (hasPercussionTrack)
                        {
                            OnFail(startIndex, "Already has percussion track");
                            return false;
                        }

                        hasPercussionTrack = true;
                        midiChannels.Add(new MidiChannel(this, midiOut, midiString, startIndex, 9, startTime));
                    }
                    else
                    {
                        if (channel == 16)
                        {
                            OnFail(startIndex, "Only 16 channels are allowed");
                            return false;
                        }

                        midiChannels.Add(new MidiChannel(this, midiOut, midiString, startIndex, channel, startTime));
                    }
                    channel += channel == 8 ? 2 : 1;
                }
                else
                {
                    OnFail(startIndex, "Channel has no content");
                    return false;
                }

                startIndex = 1 + midiString.IndexOf('|', startIndex);
            }
            while (startIndex > 0);

            if (midiChannels.Count == 0)
            {
                OnFail(startIndex, "No channels to play");
                return false;
            }

            tmr = new DispatcherTimer();
            tmr.Interval = TimeSpan.FromMilliseconds(1);
            tmr.Tick += OnTick;
            tmr.Start();

            IsPlaying = true;

            return true;
        }

        public void Stop()
        {
            if (IsPlaying)
            {
                Cleanup();
            }
        }

        void OnTick(object sender, EventArgs e)
        {
            removeChannelList.Clear();

            foreach (MidiChannel midiChannel in midiChannels)
            {
                bool channelStillGoing = true;

                try
                {
                    channelStillGoing = midiChannel.Tick();
                }
                catch (MidiParsingException exc)
                {
                    OnFail(exc.Index, exc.Message);
                    return;
                }

                if (!channelStillGoing)
                {
                    removeChannelList.Add(midiChannel);
                }
            }

            foreach (MidiChannel midiChannel in removeChannelList)
                midiChannels.Remove(midiChannel);

            if (midiChannels.Count == 0 && shutDownTime == DateTime.MaxValue)
            {
                shutDownTime = DateTime.Now + TimeSpan.FromSeconds(1);
            }

            if (DateTime.Now > shutDownTime)
            {
                OnEnded();
            }
        }

        protected void OnEnded()
        {
            Cleanup();

            if (Ended != null)
                Ended(this, EventArgs.Empty);
        }

        protected void OnFail(int index, string error)
        {
            Cleanup();

            if (Failed != null)
                Failed(this, new MidiStringPlayerFailedEventArgs(index, error));
        }

        bool HasContent(string str, int index)
        {
            int length = str.Length;

            while (index < length && str[index] != '|')
            {
                if (!Char.IsWhiteSpace(str, index))
                    return true;

                index++;
            }

            return false;
        }

        void OnLoaded(object sender, RoutedEventArgs args)
        {
            if (PlayOnLoad)
                Play();
        }

        void OnWindowClosed(object sender, EventArgs args)
        {
            Stop();
        }

        void Cleanup()
        {
            if (midiOut != null)
            {
                midiOut.Reset();
                midiOut.Close();
                midiOut.Dispose();
                midiOut = null;
            }

            if (tmr != null && tmr.IsEnabled)
            {
                tmr.Stop();
            }

            IsPlaying = false;
        }
    }
}
